#################################################################################################
#		EXIF	
#################################################################################################
# This script is a part of the log2timeline project and provides a parser for exif information
# found inside various media files, such as JPEG,PNG,and others
# 
# Author: Kristinn Gudjonsson
# Version : 0.4
# Date : 18/03/10
#
# Copyright 2009,2010 Kristinn Gudjonsson (kristinn ( a t ) log2timeline (d o t) net)
#
#  This file is part of log2timeline.
#
#    log2timeline is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    log2timeline is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with log2timeline.  If not, see <http://www.gnu.org/licenses/>.

package Log2t::input::exif;

use strict;
use Image::ExifTool qw(:Public);
use Getopt::Long; 	# read parameters
use Log2t::Common qw/LITTLE_E BIG_E/;
use Log2t::Time;	# to manipulate time
#use Log2t::Numbers;	# to manipulate numbers
use Log2t::BinRead;	# methods to read binary files


use vars qw($VERSION);

$VERSION = '0.4';

# define all global variables needed for this input module
my %info;
my $meta;
my ($username,$hostname);
my $path;

my $debug;

#       get_version
# A simple subroutine that returns the version number of the format file
#
# @return A version number

sub get_version()
{
	return $VERSION;
}

#       get_description
# A simple subroutine that returns a string containing a description of 
# the funcionality of the format file. This string is used when a list of
# all available format files is printed out
#
# @return A string containing a description of the format file's functionality
sub get_description()
{
	return "Extract metadata information from files using ExifTool"; 
}

#       prepare_file
# This subroutine starts by .. 
# 
#
# @params One parameter is defined, the path to the Prefetch directory 
# @return An integer is returned to indicate whether the file preparation was 
#       successful or not.
sub prepare_file()
{
	my $class = shift;
	my $file = shift;

	# we need to know the file name
	$info{'file_name'} = $file;

	# initialize
	$username = 'unknown';
	$hostname = 'unknown';

	# initialize and reset the index counter
	$info{'count'} = (scalar keys %info) - 2;
	$info{'index'} = 0;

        # read options from CMD
        @ARGV = @_;
        GetOptions(
                "user=s"=>\$username,
		"host=s"=>\$hostname,
		"path=s"=>\$path,
		"debug!"=>\$debug
        );

	return 1;
}

#       load_line
# A subroutine that reads a line from the access file and returns it to the
# main script
# @return A string containing one line of the log file (or a -1 if we've reached 
#       the end of the log file)
sub load_line()
{
	# check if we have reached the end
	return 0 unless  $info{'index'} < $info{'count'};

	# not reached the end, so let's continue

	# just to check if this is the file modification
	if( $info{$info{'index'}}->{'key'} eq 'FileModifyDate' )
	{
		# we don't care about the file mtime of the file
		# since that is already displayed in the filesystem metadata
		$info{'index'}++;
		return load_line();
	}
	return 1;

}
#       close_file
# A subroutine that closes the file, after it has been parsed
# @return An integer indicating that the close operation was successful
sub close_file()
{
	return 1;
}

#       parse_line
# This is the main "juice" of the format file.  It takes a line from the log file
# and parses it to produce an array containing all the needed values to print a 
# body file.
# 
# The default structure of Squid log file is:
# timestamp elapsed IP/Client Action/Code Size Method URI Ident Hierarchy/From Content
#
# @param LINE a string containing a single line from the access file
# @return Returns a array containing the needed values to print a body file

sub parse_line()
{
	# the timestamp object
	my %t_line;
	my $value = $info{$info{'index'}}->{'key'};

	#if ( $meta->{'FileType'} =~ m/EXE|DLL/ )	# unable to do since we do not receive File tags
	if ( ( $info{'file_name'} =~ m/\.dll|exe|sys/i ) )# || ( $info{'file_name'} =~ m/\.exe/i ) )
	{
		# we have an executable, let's check if we have the timestamp "TimeStamp"
		if ( lc( $value ) eq 'timestamp' )
		{
			# modify the timestamp value to a more descriptive value
			$value = 'PE header TimeDate Stamp (when application was linked/compiled)';
		}
		else
		{
			# we are still left with a PE header, so let's make it more descriptive
			$value = 'PE Header: ' . $info{$info{'index'}}->{'key'};
		}
	}

        # content of array t_line ([optional])
        # %t_line {        #       time
        #               index
        #                       value
        #                       type
        #                       legacy
        #       desc
        #       short
        #       source
        #       sourcetype
        #       version
        #       [notes]
        #       extra
        #               [filename]
        #               [md5]
        #               [mode]
        #               [host]
        #               [user]
        #               [url]
        #               [size]
        #               [...]
        # }

        # create the t_line variable
        %t_line = (
                'time' => { 0 => { 'value' => $info{$info{'index'}}->{'time'}, 'type' => $info{$info{'index'}}->{'group'} . '/' . $value, 'legacy' => 15 } },
                'desc' => $value,
                'short' => $value,
                'source' => 'EXIF',
                'sourcetype' => 'EXIF metadata',
                'version' => 2,
        );

	# we sometimes get empty lines, check for that (and ignore)
	$t_line{'desc'} = '' unless defined $info{$info{'index'}}->{'time'};

	$info{'index'}++;
	return \%t_line;
}

#       get_help
# A simple subroutine that returns a string containing the help 
# message for this particular format file.
# @return A string containing a help file for this format file
sub get_help()
{
	return "This format file parses files to search for metadata information embedded in the file.  All metadata information that contains date values is then processes.
The format file extracts all metadata information that the library Image::ExifTool is capable of reading.  That means that not only JPG or other image formats can be read, ExifTool is capable of reading metadata from various sources.";

}

#       verify
# A subroutine that verifies if we are examining a prefetch directory so it can be further 
# processed.  The correct format is a directory that consists of a folder that contains
# several files that end with a .pf ending.  Then one file in the folder is named Layout.ini
# @return An array containing an integer and a string.  The integer indicates a success or failure and the
#       string is the error message (if the file is not correctly formed)
sub verify($$)
{
	# define an array to keep
	my %return;
	my $class = shift;
	my $file = shift;
	my $detail = shift;
	my @tags;

	my $exif;

	# this needs to be a file to be included
	return \%return unless -f $$file;

	# create an exif object
	$exif = new Image::ExifTool;

	$return{'success'} = 0;
	$return{'msg'} = 'unable to open file';

	#print STDERR "[EXIF] About to open file $file\n";
	# create an image handle
	open( FH, $$file ) or return \%return;
	#print STDERR "[EXIF] Opened the file in question\n";

	# check if this is a XML file, we don't want to spend time parsing them
	unless( $detail )	
	{

		# check the name (does it end with a .xml)
		return \%return if $$file =~ m/\.xml$/i;

		# check to see if it is a XML file nonetheless
		binmode(FH);

	        # start by setting the endian correctly
	        Log2t::BinRead::set_endian( LITTLE_E );

		# read the temporary value
		my $ofs = 0;
		my $temp = Log2t::BinRead::read_ascii_until( \*FH, \$ofs, "\n", 10 );

		# now to close the file
		close(FH);

		# we don't want to parse XML files (even if they are called manifest or xib)
		if( lc($temp) =~ m/<\?xml/ )
		{
			$return{'msg'} = 'Parsing XML files using ExifTools takes too long time';
			return \%return;
		}

		# and now to open the file again, since we "passed" the first test
		open( FH, $$file ) or return \%return;
	}

	# set the date variable
	$exif->Options( DateFormat => "%Y:%m:%d %H:%M:%S" );
	# processing would be considerably easier using Epoch from the gecko
	#$exif->Options( DateFormat => "%s" );
	
	# we don't care about filesystem timestamps
	$exif->Options( Group0 => [ '-File', '-ZIP' ] );

	# reset the info hash (since we could be recursively going through)
	%info = undef;

	# default values
	$return{'success'} = 0;
	$return{'msg'} = 'Did not find any valid dates (except filesystem modification date)';


	# try to get metadata information from the file
	eval
	{
		#print STDERR "[EXIF] Trying to read metadata information from file\n";
		$meta = $exif->ImageInfo( \*FH );
		#print STDERR "[EXIF] Successfully read the metadata, now moving on to parsing it\n";
		close( FH );
	};
	if ( $@ )
	{
		close( FH );
		$return{'success'} = 0;
		$return{'msg'} = "Unable to obtain metadata information";
	}

	# now we need to verify that meta is of the correct format
	# since we can read metadata information from all files, this will
	# always return 1, even though the file will not produce any valid data

	# we therefore start by constructing the file for later use ( instead of 
	# using the prepare function )
	# since we want to know if there are any time values that we can use

	# set the counter up
	$info{'index'} = 0;

	# get all the tags (just interested in Group0 information
	@tags = $exif->GetTagList($meta,'Group0');

	# find all dates inside the meta
	foreach( @tags ) 
	{
		#print STDERR "[EXIF] Parsing tag $_\n";

		# verify that we have a valid date
		if( $meta->{$_} =~ m/\d+:\d+:\d+/ )
		{
			#print STDERR "[EXIF] Perhaps a date\n";

			# try to parse
			if( Log2t::Time::is_date( $meta->{$_} ) )
			{
				#print STDERR "[EXIF] We have a date ($_) " . $meta->{$_} . "\n";

				$info{$info{'index'}} = {
					'time' => Log2t::Time::exif_to_epoch( $meta->{$_} ),
					'key' => $_,
					'org' => $meta->{$_},
					'group' => $exif->GetGroup( $_, 0 )
				};
				$info{'index'}++;
			}
		}		
	}


	# now we've got the structure up and running, let's find out
	# if we have a valid file that we would like to examine
	$return{'success'} = 1 if $info{'index'} > 0;

	return \%return;
}

1;
