#################################################################################################
#		FIREFOX3	
#################################################################################################
# this script reads the places.sqlite SQLITE database that contains information about Firefox
# history files (3.0+) and produces a bodyfile containing the timeline information
# according to the output file used by log2timeline
#
# Database information was both found using sqlite directly as well from the web
# site : http://www.firefoxforensics.com/
#
# For ubuntu, please install
#	apt-get install libdbi-perl
# 
# Author: Kristinn Gudjonsson
# Version : 0.8
# Date : 18/03/10
#
# Copyright 2009-2010 Kristinn Gudjonsson (kristinn ( a t ) log2timeline (d o t) net)
#
#  This file is part of log2timeline.
#
#    log2timeline is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    log2timeline is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with log2timeline.  If not, see <http://www.gnu.org/licenses/>.
#
# This script uses part of the code ff3histview, previously published
# by the same author
package Log2t::input::firefox3;

use strict;
use DBI;
use File::Copy;
use Getopt::Long; # read parameters
use Log2t::Numbers;
use Log2t::Time;
use Log2t::Common;

use vars qw($VERSION);

$VERSION = '0.8';

my @r_line;
my $r_type;
my %records;
my $index;
my $username;
my $o_host;
my ($db_lock,$db_file,$db);
	
my %type = (
	'1' => 'LINK',
	'2' => 'TYPED',
	'3' => 'BOOKMARK',
	'4' => 'EMBED',
	'5' => 'REDIRECT_PERMANENT',
	'6' => 'REDIRECT_TEMPORARY',
	'7' => 'DOWNLOAD'
);

#       get_version
# A simple subroutine that returns the version number of the format file
#
# @return A version number
sub get_version()
{
	return $VERSION;
}

#       get_description
# A simple subroutine that returns a string containing a description of 
# the funcionality of the format file. This string is used when a list of
# all available format files is printed out
#
# @return A string containing a description of the format file's functionality
sub get_description()
{
	return "Parse the content of a Firefox 3 history file"; 
}

#       prepare_file
# This subroutine starts by .. 
# 
#
# @params One parameter is defined, the path to the Prefetch directory 
# @return An integer is returned to indicate whether the file preparation was 
#       successful or not.
sub prepare_file()
{
	# define needed variables
	my $temp;
	my $return = 1;
	my ($sth, $sql);
	my $result;
	my @dump;
	my $i = 0;
	my $path;

	my $class = shift;
	$db_file = shift;

	# find out username, if defined
	$username = 'unknown';
	$o_host = 'unknown';

	# read options from CMD
	@ARGV = @_;
	GetOptions(
        	"user:s"=>\$username,
		"path:s"=>\$path,
		"host:s"=>\$o_host
	);

	# set the db_lock variable
	$db_lock = 0;
	# initialize the index variable
	$index = 0;

	# check if we need to "guess" the username of the user
	if( $username eq 'unknown' )	
	{
		$username = Log2t::Common::get_username_from_path($db_file);
	}

	
	# start by checking if we have a database journal as well
	if ( -e "$db_file-journal" )
	{
		# create a new variable to store the temp location
		my $temp = rand();
		$temp = "/tmp/tmp_ff.$temp.db";

		print STDERR "Database is locked:\nDatabase copied to a temporary location to work with unlocked copy\n";
		# we need to copy the file to a temp location and start again
		copy( $db_file, $temp ) || ($return = 0);
		copy( $db_file . "-journal", $temp . "-journal" ) || ($return = 0);

		#print STDERR "[Firefox] Created a temp file $temp from $db_file\n";

		if( !$return )
		{
			print STDERR "[Firefox3] Unable to copy database file to a temp location, $temp\n";
			return $return;
		}

		$db_file = $temp;
		$db_lock = 1;
	}
	
	# connect to the database
	$db = DBI->connect("dbi:SQLite:dbname=$db_file","","") || ($return = 0);
	if( !$return )
	{
		print STDERR "[Firefox 3] Unable to connect to database\n";
		return $return;
	}
	
	# check if this is real Firefox database
	$db->prepare( 'SELECT id FROM moz_places LIMIT 1' ) || ($return = 0);

	if( !$return )
	{
		print STDERR "[Firefox 3] The database is not a correct Firefox database\n";
		return $return;
	}

	# we now know that we have a FireFox SQLITE database, let's continue

	##################################################################
	# the structure/schema of the moz_places table
	##################################################################
	# id INTEGER PRIMARY KEY
	# url LONGVARCHAR
	# title LONGVARCHAR
	# rev_host LONGVARCHAR
	# visit_count INTEGER DEFAULT 0
	# hidden INTEGER DEFAULT 0 NOT NULL, 
	# typed INTEGER DEFAULT 0 NOT NULL, 
	# favicon_id INTEGER, 
	# frecency INTEGER DEFAULT -1 NOT NULL
	# structure of moz_historyvisits
	# id INTEGER PRIMARY KEY
	# from_visit INTEGER
	# place_id INTEGER
	# visit_date INTEGER
	# visit_type INTEGER
	# session INTEGER
	#-----------------------------------------------------------------

	# Construct the SQL statement to extract the needed data
	$sql = "
SELECT moz_historyvisits.id, moz_places.url,moz_places.title,moz_places.visit_count,moz_historyvisits.visit_date,moz_historyvisits.from_visit,moz_places.rev_host,moz_places.hidden,moz_places.typed,moz_historyvisits.visit_type
FROM moz_places, moz_historyvisits
WHERE
	moz_places.id = moz_historyvisits.place_id
	";

	$sth = $db->prepare( $sql );
	$result = $sth->execute( );

	# load the result into an array
	while( @dump = $sth->fetchrow_array() )
	{
		# insert into the hash
                $records{$i} = {
                        'type' => 'url',
                        'data' => [@dump]
                };

		$i++;
	}

	# extract bookmark information 
	$sql = "
SELECT moz_bookmarks.type,moz_bookmarks.title,moz_bookmarks.dateAdded,moz_bookmarks.lastModified,moz_places.url,moz_places.title,moz_places.rev_host,moz_places.visit_count
FROM moz_places, moz_bookmarks
WHERE
	moz_bookmarks.fk = moz_places.id
	AND moz_bookmarks.type <> 3
	";

	$sth = $db->prepare( $sql );
	$result = $sth->execute( );

	# load the result into an array
	while( @dump = $sth->fetchrow_array() )
	{
		# insert into the hash
                $records{$i} = {
                        'type' => 'bookmark',
                        'data' => [@dump]
                };

		$i++;
	}

	$sql = "
SELECT moz_items_annos.content, moz_items_annos.dateAdded,moz_items_annos.lastModified,moz_bookmarks.title, moz_places.url,moz_places.rev_host
FROM moz_items_annos,moz_bookmarks,moz_places
WHERE
	moz_items_annos.item_id = moz_bookmarks.id
	AND moz_bookmarks.fk = moz_places.id
	";

	$sth = $db->prepare( $sql );
	$result = $sth->execute( );

	# load the result into an array
	while( @dump = $sth->fetchrow_array() )
	{
		# insert into the hash
                $records{$i} = {
                        'type' => 'annos',
                        'data' => [@dump]
                };

		$i++;
	}

	$sql = "
SELECT moz_bookmarks.title,moz_bookmarks.dateAdded,moz_bookmarks.lastModified
FROM moz_bookmarks
WHERE
	moz_bookmarks.type = 2
	";

	$sth = $db->prepare( $sql );
	$result = $sth->execute( );

	# load the result into an array
	while( @dump = $sth->fetchrow_array() )
	{
		# insert into the hash
                $records{$i} = {
                        'type' => 'bookmark_folder',
                        'data' => [@dump]
                };

		$i++;
	}

	return $return;;
}

#       load_line
# A subroutine that reads a line from the access file and returns it to the
# main script
# @return A string containing one line of the log file (or a -1 if we've reached 
#       the end of the log file)
sub load_line()
{
	my $t;

	if( $index < keys( %records ) )
	{
		#@r_line = @{$records{$index}};
                @r_line = @{$records{$index}->{'data'}};
                $r_type = $records{$index}->{'type'};

		$index++;

		return 1;
	}
	else
	{
		return 0;
	}
}

#       close_file
# A subroutine that closes the file, after it has been parsed
# @return An integer indicating that the close operation was successful
sub close_file()
{

	# disconnect the database
	$db->disconnect if defined $db;

	# check to see if database is locked, then delete temp file
	if( $db_lock )
	{
		#print STDERR "Deleting temporary history file, $db_file \n";
		unlink( $db_file );
		unlink( $db_file . '-journal' ) if -e $db_file . '-journal';
	}

}

#       parse_line
# This is the main "juice" of the format file.  It takes a line from the log file
# and parses it to produce an array containing all the needed values to print a 
# body file.
# 
# The default structure of Squid log file is:
# timestamp elapsed IP/Client Action/Code Size Method URI Ident Hierarchy/From Content
#
# @param LINE a string containing a single line from the access file
# @return Returns a array containing the needed values to print a body file

sub parse_line()
{
	# the timestamp object
	my %t_line;

	# fields of interest
	my ($adate,$cdate);
	my $time_offset = 0;
	my $hostname;
	my $from;
	my $text;
	my $title;
	my $type;

	# start by check which table we are examining
	if( $r_type eq 'url' )
	{
		$type = 'URL visited';

		# insert each item in line into the array 
		# inside: 
		#	moz_historyvisits.id
		#	url,
		#	title,
		#	visit_count,
		#	visit_date,
		#	from_visit,
		#	rev_host,
		#	hidden
		#	moz_places.typed,
		#	moz_historyvisits.visit_type
	
		# we need to fix the date, so we can represent it correctly (it is stored in UTC)
		$adate = Log2t::Numbers::roundup( $r_line[4] / 1000000 );
		# only one timestamp here
		$cdate = $adate;

		# get the hostname	
		$hostname = fix_hostname( $r_line[6] );

		# check to see if we came to the web site from another one
		if( $r_line[5] ne 0 )
		{
			# FROM -> URL HOST DATE
			$from = get_url( $r_line[5] );
		}

		if( $from->{'s_url'} ne '' )
		{
			$text .=  $r_line[1] . ' (' . $r_line[2] . ') [count: ' . $r_line[3] . '] Host: ' . $hostname . ' visited from: ' . $from->{'s_url'};
		}
		else
		{
			$text .= $r_line[1] . ' (' . $r_line[2] . ') [count: ' . $r_line[3] . '] Host: ' . $hostname;
		}

		if( $r_line[7] eq 1 )
		{
			$text .= ' (url hidden)';
		}
		if( $r_line[8] eq 1 )
		{
			$text .= ' (directly typed)';
		}
		else
		{
			$text .= ' (URL not typed directly)';
		}

		# and include the visit type
		$text .= ' type: ' . $type{$r_line[9]};

		$title = 'URL: ' . $r_line[1];
	}
	elsif( $r_type eq 'bookmark' )
	{
		$type = 'bookmark';
		# The array r_line now contains:
		#	0 moz_bookmarks.type
		#	1 moz_bookmarks.title
		#	2 moz_bookmarks.dateAdded
		#	3 moz_bookmarks.lastModified
		#	4 moz_places.url
		#	5 moz_places.title
		#	6 moz_places.rev_host
		#	7 moz_places.visit_count
	
		# start by extrating data
		$hostname = fix_hostname( $r_line[6] );

		# extract the dateAdded
		$cdate = Log2t::Numbers::roundup( $r_line[2] / 1000000 );

		# extract the lastModified 
		$adate = Log2t::Numbers::roundup( $r_line[3] / 1000000 );

		# now to check if one is zero
		if( ( $adate == 0 ) && ( $cdate == 0 ) )
		{
			# both dates are 0
			%t_line = '' ;
			return \%t_line;
		}
		else
		{
			if( $adate == 0 )
			{
				# then cdate isn't
				$adate = $cdate;
			}
			if( $cdate == 0 )
			{
				# then adate isn't
				$cdate = $adate;
			}
		}

		# now we need to construct the text variable
		# moz_bookmarks.type can be one of three things
		#	1 => Bookmark (aka URL)
		#	2 => Folder
		#	3 => Separator
		# we should only be getting type one since we constructed the SQL statement to get the URL as well
		if( $r_line[0] eq 1 )
		{
			$text .= ' Bookmark URL ' . $r_line[1] . ' (' . $r_line[4] . ') [' . $r_line[5] .'] count ' . $r_line[7];
			$title = ' Bookmarked ' . $r_line[1] . ' (' . $r_line[4] . ')';
		}
		else
		{
			# we don't want that
			$text = '';
			$title = '';
		}

	}
	elsif ( $r_type eq 'bookmark_folder' )
	{
		$type = 'bookmark folder';
		# The array r_line now contains:
		#	0 title
		#	1 dateAdded
		#	2 lastModified

		# extract the dateAdded
		$cdate = Log2t::Numbers::roundup( $r_line[1] / 1000000 );

		# extract the lastModified 
		$adate = Log2t::Numbers::roundup( $r_line[2] / 1000000 );

		# now to check if one is zero
		if( ( $adate == 0 ) && ( $cdate == 0 ) )
		{
			# both dates are 0
			%t_line = '' ;
			return \%t_line;
		}
		else
		{
			if( $adate == 0 )
			{
				# then cdate isn't
				$adate = $cdate;
			}
			if( $cdate == 0 )
			{
				# then adate isn't
				$cdate = $adate;
			}
		}

		$text .= 'Bookmark Folder [' . $r_line[0] . ']';
		$title = 'Bookmark folder ' . $r_line[0];
	}
	elsif( $r_type eq 'annos' )
	{
		$type = 'annotations';

		# The array r_line now contains:
		#	0 moz_items_annos.content
		#	1 moz_items_annos.dateAdded
		#	2 moz_items_annos.lastModified
		#	3 moz_bookmarks.title
		#	4 moz_places.url
		#	5 moz_places.rev_host

		# extract the dateAdded
		$cdate = Log2t::Numbers::roundup( $r_line[1] / 1000000 );

		# extract the lastModified 
		$adate = Log2t::Numbers::roundup( $r_line[2] / 1000000 );

		# now to check if one is zero
		if( ( $adate == 0 ) && ( $cdate == 0 ) )
		{
			# both dates are 0
			%t_line = '' ;
			return \%t_line;
		}
		else
		{
			if( $adate == 0 )
			{
				# then cdate isn't
				$adate = $cdate;
			}
			if( $cdate == 0 )
			{
				# then adate isn't
				$cdate = $adate;
			}
		}

		$text .= 'Bookmark Annotation: [' . $r_line[0] . '] to bookmark [' . $r_line[3] . '] (' . $r_line[4] . ')';
		$text =~ s/\n//g;
		$text =~ s/\r//g;
		$title = 'Bookmark Annotation to bookmark ' . $r_line[3] . ' (' . $r_line[4] .')';
	}
	else
	{
		# unknown type
		$text = '';
		$adate = 0;
		$cdate = 0;
		$title = '';
	}

        # content of array t_line ([optional])
        # %t_line {        #       time
        #               index
        #                       value
        #                       type
        #                       legacy
        #       desc
        #       short
        #       source
        #       sourcetype
        #       version
        #       [notes]
        #       extra
        #               [filename]
        #               [md5]
        #               [mode]
        #               [host]
        #               [user]
        #               [url]
        #               [size]
        #               [...]
        # }

        # create the t_line variable
        %t_line = (
                'desc' => $text,
                'short' => $title,
                'source' => 'WEBHIST',
                'sourcetype' => 'Firefox 3 history',
                'version' => 2,
                'extra' => { 'user' => $username, 'host' => $o_host }
        );

	# now we need to add the timestamp (check which type this is first)
	if( $r_type eq 'url' )
	{
		# url, one date
		$t_line{'time'}->{0} = { 'value' => $adate, 'type' => 'URL visited', 'legacy' => 15 };
	}
	else
	{
		# add the dates
		$t_line{'time'}->{0} = { 'value' => $adate, 'type' => 'LastModified', 'legacy' => 12 };
		$t_line{'time'}->{1} = { 'value' => $cdate, 'type' => 'dateAdded', 'legacy' => 3 };
	}

	return \%t_line;
}

#       get_help
# A simple subroutine that returns a string containing the help 
# message for this particular format file.
# @return A string containing a help file for this format file
sub get_help()
{
	return "This plugin parses the content of the places.sqlite file, which is a SQLITE 
database that contains all the information about history visits in Firefox, version 3 and newer.
places.sqlite file is usually found at the following location:
        [win xp] c:\\Documents and Settings\\USER\\Application Data\\Mozilla\\Firefox\\Profiles\\PROFILE\\places.sqlite
        [linux] /home/USER/.mozilla/firefox/PROFILE/places.sqlite
        [mac os x] /Users/USER/Library/Application Support/Firefox/Profiles/PROFILE/places.sqlite\n

There are two options to this script
	-u USERNAME
	-h HOSTNAME
Which is used to define the username of the user that ownes the history file or the host name of the machine in question\n";

}

#       verify
# A subroutine that verifies if we are examining a prefetch directory so it can be further 
# processed.  The correct format is a directory that consists of a folder that contains
# several files that end with a .pf ending.  Then one file in the folder is named Layout.ini
# @return An array containing an integer and a string.  The integer indicates a success or failure and the
#       string is the error message (if the file is not correctly formed)
sub verify($$)
{
	# define an array to keep
	my %return;
	my $line;
	my @words;
	my $temp;
 	my $magic = 'SQLite format 3';
	my $class = shift;
	my $file = shift;
	my $detail = shift;
	my ($vdb,$vsth);
	my $file_lock = 0;
	my $db;
	
	# we assume that the file is not FF3
	$return{'success'} = 0;
	$return{'msg'} = 'unknown';

	return \%return unless -f $$file;

	# open the file (at least try to open it)
	eval
	{
		open( FILE, $$file );
		binmode(FILE);

	};
	if( $@ )
	{
		$return{'success'} = 0;
		$return{'msg'} = "Unable to open file";
	}

	# check if this is a quick test (default) or a detailed one
	#unless( $detail )
	#{	
	#	# check the first bit, it has to match the magic value before proceeding
	#	seek(FILE,0,0);
	#	read(FILE,$temp,1);
	#
	#	$return{'msg'} = 'Wrong magic value';
	#	close( FILE ) unless $temp eq 'S';
	#	return \%return unless $temp eq 'S';
	#}

	# read the first few bits, see if it matches signature
	for( my $i=0 ; $i < 15; $i++ )
	{
		seek(FILE,$i,0);
		read(FILE,$temp,1);
		push( @words, $temp );
	}

	# create a string from the read characters
	$line = join('',@words);

	# close the file (for binary reading)
	close( FILE );

	# check if the line indicates a real sqlite database (version 3)
	if( $line eq $magic )
	{
		# we know that this is a SQLite database, but is it a Firefox3 database?
		$db = $$file;
	
		# start by checking if we have a database journal as well
		if ( -f $db . "-journal" )
		{
			eval
			{
				my $msg = undef;

				# create a new variable to store the temp location
				$temp = rand();
				$temp = '/tmp/tmp_ff.' . $temp . 'v.db';
			
				$return{'msg'} = 'Unable to copy the file to a temporary folder';
				$return{'success'} = 0;

				# we need to copy the file to a temp location and start again
				copy( $db, $temp ) or $msg = 'unable to copy database file. ';
				copy( $db . "-journal", $temp . "-journal" ) or $msg .=  'unable to copy journal file to temporary directory'; 

				# we are trying to copy temporary data, and if the msg variable is defined, then we have a problem
				if( defined $msg )
				{
					$return{'msg'} = 'Error while trying to verify, unable to verify.  The error is: ' . $msg;
					$return{'success'} = 0;
					return \%return;
				}
				#print STDERR "[Firefox] Created a temp file $temp from $db\n";
		
				$db = $temp;
				$file_lock = 1;	# indicate that we need to delete the lock file
			};
			if( $@ )
			{
				$return{'success'} = 0;
				$return{'msg'} = 'Database is locked and unable to copy to a temporary location (' . $temp . '). Error given: ' . $@;


				return \%return;
			}
		}	

		# set a temp variable to 0 (assume we don't have a FF3.5 database)
		$temp = 0;
	
		eval
		{
			# connect to the database
			$vsth = $vdb = DBI->connect("dbi:SQLite:dbname=$db","","") or ( $temp = 1);

			if( $temp )
			{
				$return{'success'} = 0;
				$return{'msg'} = 'Unable to connect to the database';
				return \%return;
			}
	
			# check if this is real Firefox database
			#$vsth = $vdb->prepare( 'SELECT id FROM moz_places LIMIT 1' ) or ($temp = 0);

			# get a list of all available talbes
			$vsth = $vdb->prepare( "SELECT name FROM sqlite_master WHERE type='table'" ) or ($temp = 0);

			#execute the query
			$temp = 0;
			my $res = $vsth->execute();

			# check if we have a moz_places table
		        while( @words = $vsth->fetchrow_array() )
		        {
				# check for moz_places
				#print STDERR "RESULT IS " . $words[0] . "\n";
				$temp = 1 if $words[0] eq 'moz_places';
        		}

			if( $temp )
			{
				# now we have a FF3+ history SQLite database
				$return{'success'} = 1;
				$return{'msg'} = 'Success';
			}
			else
			{
				$return{'success'} = 0;
				$return{'msg'} = 'This is not a Firefox 3 history SQLite database';
			}
		
			# disconnect from the database
			$vsth->finish;
			undef $vsth;
			$vdb->disconnect;

			# delete the temporary database file (if it was locked)
			if( $file_lock ) 
			{
				#print STDERR "[Firefox] Deleting temporary files\n";
				unlink( $db ) or print STDERR "[Firefox] Unable to delete the temporary database file $db [$!] \n";

				if( -f $db . '-journal' )
				{
					unlink( $db . '-journal' ) or print STDERR "[Firefox] Unable to delete the temporary database file " . $db . "-journal (journal) [$!]\n";
				}
				#else
				#{
				#	print STDERR "[Firefox] No journal file\n";
				#}

			}
		};
		if( $@ )
		{
			$return{'success'} = 0;
			$return{'msg'} = 'Database error ocurred, making verification not possible.  The error message is: ' . $@;
		}
	}
	else
	{
		$return{'success'} = 0;
		$return{'msg'}  = "Wrong magic value.  Is this really a sqlite database?\n";
	}
	
	return \%return;
}


# functions

#	get_url
# This function takes as an input a ID from the table moz_historyvisists and
# returns a simple array containing few of relevant information from that URL
#
# @param id	The identification number for the URL in the moz_historyvisists table
# @return	An array containing the hostname, URL and date of visit
sub get_url
{
	my $statement;
	my %return;
	my $sql;
	my $result;
	
	# construct the SQL statement
	$sql = "
SELECT url,rev_host,visit_date 
FROM moz_places, moz_historyvisits
WHERE
	moz_places.id = moz_historyvisits.place_id
	AND moz_historyvisits.id = ?
	";

	$statement = $db->prepare( $sql );
	$result  = $statement->execute( $_[0] );
	
	# retrieve the results
	($return{'s_url'}, $return{'s_host'}, $return{'s_date'} ) = $statement->fetchrow_array();

	# fix variables
	$return{'s_host'} = fix_hostname( $return{'s_host'} );	
	$return{'s_date'} = Log2t::Numbers::roundup( $return{'s_date'} / 1000000 );
	
	# return the array
	return \%return;
}

#	fix_hostname
# This function takes the hostname variable, as represented in the 
# moz_places table which is in a reverse format, prepended with a dot
# and reverses it and removes the front dot (.)
#
# @params hostname A string that is of the Mozilla format for rev_host
# @return Returns a string containing the more readable version of the hostname
sub fix_hostname
{
	my $host;

	$host = reverse $_[0];
	if( $host =~ m/^\./ )
	{
		$host = substr( $host, 1, length( $host) - 1 );
	}

	return $host;
}


1;
