#################################################################################################
#		OXML	
#################################################################################################
# this script reads a log file and produces a bodyfile containing the timeline information
# that can be used directly with the script mactime from TSK collection.  
# The specification of the body file can be found here:
#	http://wiki.sleuthkit.org/index.php?title=Body_file
# This script reads a document that is written in the OpenXML format, such
# as Microsoft Office documents and displays the metadata information that
# are contained in it according to the documentation that Microsoft provides
#
# See further information about the structure here:
# 	http://msdn.microsoft.com/en-us/library/aa338205.aspx
#
# This script requires both Archive::Zip and LibXML, 
# to install the dependencies using Ubuntu, issue the following commands:
# 	apt-get install libarchive-any-perl
# 	apt-get install libxml-libxml-perl
#
# According to the standard OpenXML documents are compressed using ZIP
# and therefore it is required to unzip the documents that contain the
# metadata information before processing them further.  
# The metadata is then stored in a XML documents.  The file
# _rels/.rels defines the relationships that the document contains
# and therefore it should be the first file that is to be read.
# From there you can find any additional files that contain metadata
# information, most files will contain two metadata information files:
#	DOC.ENDING/docProps/app.xml
#	DOC.ENDING/docProps/core.xml	
#
# This script will read the _rels/.rels file, parse it's input, search
# for any XML file that contains property information of the file and 
# then parse that document and print out the metadata information found
# in it.
# 
# Author: Kristinn Gudjonsson
# Version : 0.4
# Date : 19/03/10
#
# Copyright 2009-2010 Kristinn Gudjonsson (kristinn ( a t ) log2timeline (d o t) net)
#
#  This file is part of log2timeline.
#
#    log2timeline is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    log2timeline is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with log2timeline.  If not, see <http://www.gnu.org/licenses/>.

package Log2t::input::oxml;

# check to see if the script is running on Win or *NIX 
#BEGIN{ if ( $^O =~ /MSWin32/ ) { use Win32::API;} };

use strict;

use Log2t::Common ':binary';
use Log2t::Time;
use Log2t::BinRead;
use Getopt::Long; # read parameters
use Archive::Zip qw( :ERROR_CODES );
use XML::LibXML;
use XML::LibXML::Common;
use Encode;

use vars qw($VERSION);

# version number
$VERSION = '0.4';

my ($host,$username);
my %lines;
my $index;
my @keys;
my $done;
my $key;
my $path;
my $debug;


my $encoding;
my $xml;
my $zip;
my %d_prop;
my $file;

#       get_version
# A simple subroutine that returns the version number of the format file
#
# @return A version number

sub get_version()
{
	return  $VERSION;
}

#       get_description
# A simple subroutine that returns a string containing a description of 
# the funcionality of the format file. This string is used when a list of
# all available format files is printed out
#
# @return A string containing a description of the format file's functionality
sub get_description()
{
	return "Parse the content of an OpenXML document (Office 2007 documents)"; 
}

#       prepare_file
# This subroutine starts by .. 
# 
#
# @params One parameter is defined, the path to the Prefetch directory 
# @return An integer is returned to indicate whether the file preparation was 
#       successful or not.
sub prepare_file()
{
	# define the needed variables
	my $doc;
	my $status;
	my $metadata;
	my $property;
	my $propertylist;
	my @properties;
	my @relationships;
	my $relationship;
	my $parent;
	my @list;
	my $attrib;
	my @targets;
	my $return = 1;
	my $class = shift;
	$file = shift;
	my $error = undef;

	# initialize some variables
	$host = 'unknown';
	$username = 'unknown';
	$path = undef;
	$index = 0;
	$done = 0;
	$debug = 0;
	%lines = undef;
	@keys = undef;	# initialize

        # read options from CMD
        @ARGV = @_;
        GetOptions(
                "user:s"=>\$username,
		"path:s"=>\$path,
		"debug!"=>\$debug,
		"host:s"=>\$host
        );
	
	($d_prop{dev},$d_prop{inode},$d_prop{mode},$d_prop{nlink},$d_prop{uid},$d_prop{gid},$d_prop{rdev},$d_prop{size},$d_prop{atime},$d_prop{mtime},$d_prop{ctime},$d_prop{blksize},$d_prop{blocks}) = stat($file);

	# create a ZIP object
	$zip = Archive::Zip->new();

	# read the Word document, that is the ZIP file
	$zip->read( $file ) == AZ_OK or $error =  "Unable to open Office file\n";

	if( defined $error )
	{
		print STDERR "[OXML] Error while trying to open file.  Message: $error\n";
		return 0;
	}

	# extract document information
	$status = $zip->extractMember( '_rels/.rels', '/tmp/rels.xml' );
	print STDERR "Unable to extract schema from file [$file], is it really a Office 2007 document (openXML)?\n" if $status != AZ_OK;
	return 0 if $status != AZ_OK;

	# read the rels file
	$xml = XML::LibXML->new();

	# try to read inn the relationship file
	eval
	{
		$metadata = $xml->parse_file( "/tmp/rels.xml" );
	};
	if( $@ )
	{
		# if we are unable to properly parse the relationship file, we cannot continue parsing the file, so let's gracefully exit
		print STDERR "[OXML] Unable to parse the relationship file (_rels/.rels)\n. Is this truly an OpenXML file, please verify manually.\nError message given: $@\n";
		return 0;
	}
	
	# get all the Relationship nodes
	$propertylist = $metadata->getDocumentElement();
	@properties = $propertylist->childNodes;

	# get the encoding of the document
	$encoding = $metadata->encoding();

	# examine each one
	foreach $property (@properties)
	{
		# property is a node
		if( $property->nodeType==ELEMENT_NODE )
		{
			# now we are inside the Relationship tag, find the type
			@relationships = $property->attributes();
	
			# examine each attribute that is defined for the relationshp
			foreach $relationship ( @relationships )
			{
				# we are trying to find nodes which contain property values for the file
				if( $relationship->toString =~ /.*Type.*prop.*/ )
				{
					# now we have a property that consists of a property file
					# examine each attribute that is assigned to the parent node
					$parent = $relationship->getOwnerElement();
					
					@list = $parent->attributes();
					foreach $attrib ( @list ) 
					{
						# need to find the attribute Target, since that defines
						# the location of the XML document that describes the 
						# metadata information from the document
						if ( $attrib->toString =~ /Target/ )
						{
							print STDERR "[OXML] Pushing a target, " . $attrib->value . "\n" if $debug;
							# push the name of the metadata document into the array targets
							push( @targets, $attrib->value);
						}
					}
				}
			}
		}	
	}

	# we no longer need the rels.xml file, so we delete it
	unlink( '/tmp/rels.xml' ); 

	# examine all the targets
	foreach $attrib (@targets)
	{
		# let's make an attempt of parsing the file
		eval
		{
			print STDERR "[OXML] Processing file $attrib\n" if $debug;
			unless ( process_file( $attrib ) )
			{
				print STDERR "Unable to read document metadata\n";
				return 0;
			}
		};
		if( $@ )
		{
			print STDERR "[OXML] Unable to properly parse one of the OpenXML metadata XML files.  The file $attrib was the cause of all this trouble.
Please review this file manually.  The error message given is:\n$@\n";
			return 0;
		}
	}

	foreach( keys( %lines ) )
	{
		push( @keys, $_ );
	}

	
	return $return;;
}

#       load_line
# A subroutine that reads a line from the access file and returns it to the
# main script
# @return A string containing one line of the log file (or a -1 if we've reached 
#       the end of the log file)
sub load_line()
{
	if( $index < $#keys+1  )
	{
		# check if there is a date in this line
		if( $lines{$keys[$index]} =~ m/\d{4}-\d{2}-\d{2}/ )
		{
			# we have a date
			print STDERR "[OXML] A date object " . $keys[$index] . " = [" . $lines{$keys[$index]} . "]\n" if $debug;
			$key = $keys[$index++];
			return 1;
		}
		else
		{
			$index++;
			return load_line();
		}
	}
	else
	{
		return 0;
	}
}
#       close_file
# A subroutine that closes the file, after it has been parsed
# @return An integer indicating that the close operation was successful
sub close_file()
{
	return 1;
}

#       parse_line
# This is the main "juice" of the format file.  It takes a line from the log file
# and parses it to produce an array containing all the needed values to print a 
# body file.
# 
# The default structure of Squid log file is:
# timestamp elapsed IP/Client Action/Code Size Method URI Ident Hierarchy/From Content
#
# @param LINE a string containing a single line from the access file
# @return Returns a array containing the needed values to print a body file

sub parse_line()
{
	# the timestamp object
	my %t_line;
	my $value;
	my $date;
	my $text;
	my $fname;
	
	# substitute multiple spaces with one for splitting the string into variables
	$key =~ s/\s+/ /g;

	#fix / and \ difference
	$fname = $file;
	$fname =~ s/\//\\/g;

	$value = $lines{$key};

	$date = Log2t::Time::iso2epoch( $value );

	#$text = 'File: [' . $path . '\\' . $fname . '].' if defined $path;
	#$text = 'File: [' . $fname . '].' unless defined $path;

	# construct the text field
	$text .= ' (' . $lines{title} . ')' if ( defined $lines{title} ) and ( $lines{title} ne '' );
	$text .= ' - ' . $lines{subject} . ' - ' if ( defined $lines{subject} ) and ( $lines{subject} ne '' );
	$text .= ' - ' . $lines{description} . ' - ' if ( defined $lines{description} ) and ( $lines{description} ne '' );
	$text .= ' - Application: ' . $lines{application} if defined $lines{application};
	$text .= ' - Company: ' . $lines{company} if defined $lines{company} and $lines{company} ne '';
	$text .= ' - AppVersion: ' . $lines{appversion} if defined $lines{appversion} and $lines{appversion} ne '';

	if ( exists $lines{creator} )
	{
		if( $username eq 'unknown' )
		{
			$username = $lines{creator};
		}
		elsif( $username eq $lines{creator} )
		{
			# really no action here
			print STDERR "[OXML] The username is the same as the document's creator\n" if $debug;
		}
		else
		{
			$username = $username . ' (' . $lines{creator} . ')';
		}
	}

	if ( exists $lines{description} and $lines{description} ne '' )
	{
		$text .= ' - Desc: ' . $lines{description};
	}

	$username = 'unknown' if $username eq '';
	$username = 'unknown' if $username eq ' ()';

        # content of array t_line ([optional])
        # %t_line {        #       time
        #               index
        #                       value
        #                       type
        #                       legacy
        #       desc
        #       short
        #       source
        #       sourcetype
        #       version
        #       [notes]
        #       extra
        #               [filename]
        #               [md5]
        #               [mode]
        #               [host]
        #               [user]
        #               [url]
        #               [size]
        #               [...]
        # }

        # create the t_line variable
        %t_line = (
                'time' => { 0 => { 'value' => $date, 'type' => $key, 'legacy' => 15 } },
                'desc' => $text,
                'short' => $lines{'description'},
                'source' => 'OXML',
                'sourcetype' => 'Open XML Metadata',
                'version' => 2,
                'extra' => { 'user' => $username, 'host' => $host, }
        );

	return \%t_line;
}

#       get_help
# A simple subroutine that returns a string containing the help 
# message for this particular format file.
# @return A string containing a help file for this format file
sub get_help()
{
	return "This plugin parses OpenXML documents, such as docx, pptx, xlsx and other
documents created using this standard (Office 2007 documents).  The parser relies upon the
following libraries:
	Archive::Zip 
	XML::LibXML
	XML::LibXML::Common
	Encode

It accepts two parameters:
	-h|--host HOST
		HOST is the hostname of the host that contains the document
	-u|--user USER
		USER is the user that this document belongs to\n";
}

#       verify
# A subroutine that verifies if we are examining a prefetch directory so it can be further 
# processed.  The correct format is a directory that consists of a folder that contains
# several files that end with a .pf ending.  Then one file in the folder is named Layout.ini
# @return An array containing an integer and a string.  The integer indicates a success or failure and the
#       string is the error message (if the file is not correctly formed)
sub verify($$)
{
	my $class = shift;
	my $file = shift;
	my $detail = shift;

	# define an array to keep
	my %return;
	my @words;
	my $ofs = 0;
	my %info;

	# fix the endian
	#Log2t::BinRead::set_endian( Log2t::Common::LITTLE_E );
	Log2t::BinRead::set_endian( LITTLE_E );

	# default values
	$return{'success'} = 0;
	$return{'msg'} = 'unknown';
	
	return \%return unless -f $$file;

	# open the file (at least try to open it)
	eval
	{
		open( FILE, $$file );
		binmode(FILE);

		$info{'magic'} = Log2t::BinRead::read_32(\*FILE, \$ofs );

		# check if this is truly an ZIP file (a la openXML)
		if( $info{'magic'} eq 0x04034b50 )
		{
			# then we would like to continue our reading
			$info{'version'} = Log2t::BinRead::read_16(\*FILE,\$ofs );
			$info{'general'} = Log2t::BinRead::read_16(\*FILE,\$ofs );
			$info{'comp_method'} = Log2t::BinRead::read_16(\*FILE,\$ofs );
			$info{'last_mod_time'} = Log2t::BinRead::read_16(\*FILE,\$ofs );
			$info{'last_mod_date'} = Log2t::BinRead::read_16(\*FILE,\$ofs );
		
			$info{'crc2'} = Log2t::BinRead::read_32(\*FILE,\$ofs );
			$info{'compr_size'} = Log2t::BinRead::read_32(\*FILE,\$ofs );
			$info{'size'} = Log2t::BinRead::read_32(\*FILE,\$ofs );
			$info{'filename_length'} = Log2t::BinRead::read_16(\*FILE,\$ofs );
			$info{'extra_length'} = Log2t::BinRead::read_16(\*FILE,\$ofs );
	
			$info{'filename'} = Log2t::BinRead::read_ascii(\*FILE,\$ofs,$info{'filename_length'} );

		}
		close( FILE );
	};
	if ( $@ )
	{
		$return{'success'} = 0;
		$return{'msg'} = "Unable to open file";
	}

	# 
	if( $info{'magic'} eq 0x04034b50 )
	{
		# this is an ZIP archive, now we need to confirm that this is an OpenXML document
		if( $info{'filename'} =~ m/Content_Types/ )
		{
			$return{'success'} = 1;
		}
		else
		{
			$return{'success'} = 0;
			$return{'msg'} = 'The file is a ZIP file but not an OpenXML document';
		}
	}
	else
	{
		$return{'success'} = 0;
		$return{'msg'} = 'Wrong magic value';
	}

	return \%return;
}


# ------------------------------------------------------------------------------------------------------------
#	process_file
# ------------------------------------------------------------------------------------------------------------
# This function reads a XML file that contains metadata
# information from a OpenXML file and prints out all the
# tags that are defined within it.
#
# @param xmlfile A string that contains the path within the ZIP archive that contains metadata information
# @param title A title for the file to be printed out before the metadata is printed
# @return Return false if unsuccessful, else true
sub process_file
{
	my @splits;
	my $xmlfile;
	my $status;
	my $metadata;
	my $propertylist;
	my @properties;
	my $property;

	# assign the xmlfile
	$xmlfile = shift;
	
	$status = $zip->extractMember( $xmlfile, '/tmp/file.xml' ) ;
	if ( $status != AZ_OK )
	{
		print STDERR "[OXML ERROR] Unable to extract MetaData from file, is it really a Office 2007 document?\n";
		return 0;
	}
	
	# we can now read the file
	
	# create a XML parser
	$xml = XML::LibXML->new();
	
	# read inn all the XML
	$metadata = $xml->parse_file( "/tmp/file.xml" );
	
	$propertylist = $metadata->getDocumentElement();
	@properties = $propertylist->childNodes;
	
	foreach $property (@properties)
	{
		# property is a node
		if( $property->nodeType==ELEMENT_NODE )
		{
			@splits = split( ':', $property->nodeName );

			# print the MetaData information
			if( $#splits eq 1 )
			{
				print STDERR "[OXML] Pushing key " . lc( $splits[1] ) . " into lines\n" if $debug;
				$lines{lc( $splits[1] )} = encode( $encoding, $property->textContent);
			}
			else
			{
				print STDERR "[OXML] Pushing key " . lc( $splits[0] ) . " into lines\n" if $debug;
				$lines{lc( $splits[0]) } = encode( $encoding, $property->textContent);;
			}
		
		}
	}
	
	unlink( '/tmp/file.xml' ); 
	
	return 1;
}

1;
