# ------------------------------------------------------------------------------ #
# 				prefetch
# ------------------------------------------------------------------------------ #
# This script is a part of the log2timeline program used to parse a log file and 
# output a body file that can be imported into scripts such as the mactime from the
# TSK package to provide a timeline analsysis.
#
# This script reads and parses the prefetch directory that can be found in Microsoft
# Windows (some versions) and output the content found in a body file 
#
# The specification of the body file can be found here:
#	http://wiki.sleuthkit.org/index.php?title=Body_file
#
# Author: Kristinn Gudjonsson
# Version : 0.7
# Date : 13/04/10
#
# Copyright 2009-2010 Kristinn Gudjonsson (kristinn ( a t ) log2timeline (d o t) net)
#
#  This file is part of log2timeline.
#
#    log2timeline is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    log2timeline is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with log2timeline.  If not, see <http://www.gnu.org/licenses/>.
#
# 
# Borrowed code from pref.pl written by H. Carvey
# copyright 2007 H. Carvey keydet89@yahoo.com


package Log2t::input::prefetch;

use strict;
use Getopt::Long; # read parameters
use Log2t::Time;
use Log2t::BinRead;
use Log2t::Common ':binary';
use vars qw($VERSION);

# version number
$VERSION = '0.7';

# define variables
my $directory;
my @files;
my $line;
my %exes;
my $index;
my $host;
my $vista;
my $debug;

#       get_version
# A simple subroutine that returns the version number of the format file
# @return A version number
sub get_version()
{
	return $VERSION;
}

#       get_description
# A simple subroutine that returns a string containing a description of 
# the funcionality of the format file. This string is used when a list of
# all available format files is printed out
# @return A string containing a description of the format file's functionality
sub get_description()
{
	return "Parse the content of the Prefetch directory";
}

#       prepare_file
# This subroutine starts by reading the content of the Layout.ini file, a file that 
# contains information about each of the files found in the Prefetch directory.  The
# content of the file is stored in a global array that can be used in other functions.
# 
# The script then reads all the content of the Prefetch directory and stores it in
# an array that is global.
# @params One parameter is defined, the path to the Prefetch directory 
# @return An integer is returned to indicate whether the file preparation was 
#       successful or not.
sub prepare_file()
{
	# define needed local variables
	my $layout;
	my $i;
	my $return = 1;
	my $offset;
	my $fpath;

	my $class = shift;
	$directory = shift;

	$host = 'unknown';
	$vista = 0;
	$debug = 0;
	$index = 0;
	%exes = undef;	# reset the lines arrray

        # read options from CMD
        @ARGV = @_;
        GetOptions(
		"host:s"	=>\$host,
		"debug!" => \$debug,
		"path:s"=>\$fpath,
	#	"vista!" => \$vista
        );

	# start by reading the content of the directory
	opendir( IMD, $directory ) or $return = 0;
	@files = readdir( IMD );
	closedir( IMD );

	print STDERR "[PREFETCH] The number of files inside the folder: " . $#files . "\n" if $debug;

        # open up the layout file and read it's content
        $layout = $directory . "/Layout.ini";
        open( LAYOUT, "$layout" ) or $return = 0;

	if( $debug )
	{
		print STDERR "[PREFETCH] Examining a Vista or newer operating system (superfetch)\n" if $vista;
	}
	
	while( <LAYOUT> )
	{
		# change to ascii
		s/\x00//g;
		s/\n//g;
		s/\r//g;

		if( /([a-zA-Z0-9_\-\.]+)\.EXE$/i )
		{
			$exes{lc($1)} = $_ unless $1 eq '';
		}
	}

        # close the file
        close( LAYOUT );

	return $return;
}

#       load_line
# A subroutine that reads a line from the access file and returns it to the
# main script
# @return A string containing one line of the log file (or a -1 if we've reached 
#       the end of the log file)
sub load_line()
{
	my $magic;
	my $version;
	my $ofs;

	# we've already read the content of the directory, let's output some of the files
	if ( $#files > $index-1 )
	{
		if(  ($files[$index] =~ m/^\.$/ ) or ($files[$index] =~ m/^\.\.$/ ) or ($files[$index] =~ m/Layout.ini/i ) )
		{
			print STDERR "[PREFETCH] Not parsing the file " . $files[$index] . "\n" if $debug;
			
			$index++;
			return load_line();
		}
		else
		{
			# now to check the "magic" value
			open( CF, $directory . '/' . $files[$index] );
			$ofs = 0;
			$version = Log2t::BinRead::read_16( \*CF, \$ofs );
			$ofs = 0x4;
			$magic = Log2t::BinRead::read_ascii( \*CF, \$ofs, 4 );
			close( CF );

			if( $magic eq 'SCCA' )
			{
				print STDERR "[PREFETCH] Parsing file " . $files[$index] . "\n" if $debug;

				$vista = 1 if ( $version eq 0x17 );
				$vista = 0 if ( $version eq 0x11 );

				$line = $files[$index++];
				return 1;
			}
			else
			{
				# not the correct magic value, incorrect file
				printf STDERR "[PREFETCH] Not parsing the file %s, wrong magic value (%s)\n",$files[$index],$magic if $debug;

				$index++;
				return load_line();
			}
		}
	}
	else
	{
		return 0;
	}
}

#       close_file
# A subroutine that closes the file, after it has been parsed
# @return An integer indicating that the close operation was successful
sub close_file()
{
	return 1;
}

#       parse_line
# This is the main "juice" of the format file.  It takes a line from the log file
# and parses it to produce an array containing all the needed values to print a 
# body file.
# 
# The default structure of Squid log file is:
# timestamp elapsed IP/Client Action/Code Size Method URI Ident Hierarchy/From Content
#
# @param LINE a string containing a single line from the access file
# @return Returns a array containing the needed values to print a body file
sub parse_line()
{
	# timestamp object
	my %t_line;
	my $file_name;
	my $l;
	my $path = '<path not found in Layout.ini>';
	my $runcount;
	my $runtime;
	my $exe;
	my $fpath = undef;
	my $text;

	# assign the file name variable
	$file_name = $line;

	# check the content of the Layout.ini file to find the full path of the executable
	if( $file_name =~ m/(.+)\.EXE/i )
	{
		$path = $exes{lc($1)} if defined $exes{lc($1)};
	}

	# get metadata from file
	($runcount,$runtime,$exe,$fpath) = getMetaData( "$directory/$file_name");

	# check to see if we were able to read the file
	return \%t_line unless defined $runtime;

	# now we need to fix this fpath variable
	$fpath =~ s/\\DEVICE\\HARDDISKVOLUME\d\\/\n/g;
	while( $fpath =~ m/(^.+DLL$)/gm )
	{
		$text .= $1 . ' - ';
	}
	# fix the last one
	$text =~ s/ - $//;

        # content of array t_line ([optional])
        # %t_line {        #       time
        #               index
        #                       value
        #                       type
        #                       legacy
        #       desc
        #       short
        #       source
        #       sourcetype
        #       version
        #       [notes]
        #       extra
        #               [filename]
        #               [md5]
        #               [mode]
        #               [host]
        #               [user]
        #               [url]
        #               [size]
        #               [...]
        # }

	my $st;
	$st = 'Vista/Win7' if $vista;
	$st = 'XP' unless $vista;

        # create the t_line variable
        %t_line = (
                'time' => { 0 => { 'value' => $runtime, 'type' => 'Last run', 'legacy' => 15 } },
                'desc' => $file_name . ' - [' . $exe . '] was executed - run count [' . $runcount . '], full path: [' . $path . '] - DLLs loaded: {' . $text . '}',
                'short' => $file_name . ': ' . $exe . ' was executed',
                'source' => 'PRE',
                'sourcetype' => $st . ' Prefetch',
                'version' => 2,
                'extra' => { 'host' => $host, 'filename' => "$directory/$file_name" }
        );

	# return the array that contains all the needed information to build a timeline
	return \%t_line;
}

#       get_help
# A simple subroutine that returns a string containing the help 
# message for this particular format file.
# @return A string containing a help file for this format file
sub get_help()
{
	return "This format file parses the content of the Windows prefetch directory,
%SYSTEMROOT%/Prefetch 
It both parses the .pf files it finds as well as the Layout.ini file to return values
found inside the Prefetch folder
This format file accepts the parameter
	--host	host name\n";

}

#       verify
# A subroutine that verifies if we are examining a prefetch directory so it can be further 
# processed.  The correct format is a directory that consists of a folder that contains
# several files that end with a .pf ending.  Then one file in the folder is named Layout.ini
# @return An array containing an integer and a string.  The integer indicates a success or failure and the
#       string is the error message (if the file is not correctly formed)
sub verify($$)
{
	my $class = shift;
	my $dir = shift;
	my $detail = shift;

	# define an array to keep
	my %return;
	my @files;
	my $file;
	my $line;
	my $ok_pf = 0;
	my $ok_layout = 0;

	# default values
	$return{'success'} = 0;
	$return{'msg'} = '';

	# verify if we have a folder
	return \%return unless -d $$dir;

	# start by setting the endian correctly
	Log2t::BinRead::set_endian( LITTLE_E );

	# let the fun begin
	eval
	{
		opendir( IMD, $$dir ); 
		@files = readdir ( IMD );
		closedir( IMD );
	
		# check all files in the directory
		foreach $file (@files)
		{
			if( $file =~ m/\.pf$/i )
			{
				$ok_pf = 1;
			}
	
			if( lc( $file ) eq 'layout.ini' )
			{
				$ok_layout = 1;
			}
		}
	};
	if( $@ )
	{
		$return{'msg'} = "Unable to open folder: $$dir. ";
		$return{'success'} = 0;
	}

	if( $ok_layout )
	{
		eval
		{
        		open( LAYOUT, "$$dir/Layout.ini" );
			$line = <FILE>;
			close( LAYOUT );
		};
		if ( $@ )
		{
			$return{'success'} = 0;
			$return{'msg'} .=  ' Unable to open Layout file. ';
		}
	}


	if( $ok_pf  && $ok_layout )
	{
		$return{'success'} = 1;
	}

	return \%return;
}

# 	getMetaData
# get metadata from .pf files
# copyright 2007 H. Carvey keydet89@yahoo.com
sub getMetaData 
{
	# 64
        my $file = $_[0];
        my $data;
        my ($runcount,$runtime);
	my $exe; # added by Kristinn
	my $dll_ofs;
	my $dll_length;
	my $drive_info;
	my $filepath;
	my ($volume_path,$volume_length,$volume_serial);

        open(FH,"<",$file) || return (undef,undef,undef,undef); 
        binmode(FH);

        seek(FH,0x78,0) unless $vista;
        seek(FH,0x80,0) if $vista; # if Vista/Win7

        read(FH,$data,8);
        my @tvals = unpack("VV",$data);
        $runtime = Log2t::Time::Win2Unix($tvals[0],$tvals[1]);

        seek(FH,0x90,0) unless $vista;
        seek(FH,0x98,0) if $vista; # if Vista/Win 7
        read(FH,$data,4);
        $runcount = unpack("V",$data);

	# what follows was added by Kristinn
	my $ofs = 0x10; # start of name part
	$exe = Log2t::BinRead::read_unicode_end(\*FH,\$ofs,100);
	
	# get the offset for the DLL path
	$ofs = 0x64;
	$dll_ofs = Log2t::BinRead::read_32( \*FH, \$ofs );

	#printf STDERR "[PREFETCH] OFS %s - 0x%x\n",$file, $dll_ofs;
	# get the length of the filename path
	$dll_length = Log2t::BinRead::read_32( \*FH, \$ofs );

	#printf STDERR "[PREFETCH] LENGTH 0x%x\n",$dll_length;

	#$drive_info = Log2t::BinRead::read_32( \*FH, \$ofs );
	#printf STDERR "[PREFETCH] DRIVE INFO 0x%x\n",$drive_info;

	$ofs = $dll_ofs;
	$filepath = Log2t::BinRead::read_unicode( \*FH, \$ofs, $dll_length/2 );

	# get the volume information
	#$ofs = $drive_info;
	#$volume_path = Log2t::BinRead::read_16(\*FH, \$ofs );
	#$ofs = $drive_info + 4;
	#$volume_length = Log2t::BinRead::read_32( \*FH,\$ofs );

	#printf STDERR "[PREFETCH] LENGTH 0x%x PATH 0x%x\n",$volume_length,$volume_path;

        close(FH);
        return ($runcount,$runtime,$exe, $filepath);
}

1;

