#################################################################################################
#		RECYCLER	
#################################################################################################
# this script reads the INFO2 (or I$) file/s that contains information about deleted items that
# still reside in the recycle bin of a Windows machine and produces a bodyfile containing the 
# timeline information that can be used directly with the script mactime from TSK collection.  
# The specification of the body file can be found here:
#	http://wiki.sleuthkit.org/index.php?title=Body_file
# 
# Author: Kristinn Gudjonsson
# Version : 0.5
# Date : 07/03/10
#
# Copyright 2009-2010 Kristinn Gudjonsson (kristinn ( a t ) log2timeline (d o t) net)
#
#  This file is part of log2timeline.
#
#    log2timeline is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    log2timeline is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with log2timeline.  If not, see <http://www.gnu.org/licenses/>.

package Log2t::input::recycler;

use strict;
use Log2t::Common ':binary';
use Getopt::Long; # read parameters
use Log2t::Time;
use Log2t::Numbers;
use Log2t::BinRead;
use Encode;


my $version = "0.5";
my $older_version;	# defines whether we are reading INFO2 (XP-) file or $I files (Vista +)

# define arrays to store record data
my %records;
my ($username,$host );
my $cur_line;


#       get_version
# A simple subroutine that returns the version number of the format file
#
# @return A version number
sub get_version()
{
	return $version;
}

#       get_description
# A simple subroutine that returns a string containing a description of 
# the funcionality of the format file. This string is used when a list of
# all available format files is printed out
#
# @return A string containing a description of the format file's functionality
sub get_description()
{
	return "Parse the content of the recycle bin directory"; 
}

#       prepare_file
# This subroutine starts by resetting variables and defining other needed
# variables.  It then parses the entire INFO2 (I$) file and inserts all
# the parsed values into several arrays that will be called during later
# subfunctions
#
# @params One parameter is defined, the path to the INFO2 file
# @return An integer is returned to indicate whether the file preparation was 
#       successful or not.
sub prepare_file()
{
	# define some variables
	my $class = shift;
	my $file = shift;
	my $path;

	$host = 'unknown';
	$username = 'unknown';

	$cur_line = 0;

        # read options from CMD
        @ARGV = @_;
        GetOptions(
                "user:s"	=>\$username,
		"path:s"=>\$path,
		"host:s"	=>\$host
        );


	# check if we are dealing with "older versions" of recycle bin or the later ones
	return prepare_info2($file) if $older_version;
	
	return prepare_i($file);

}

sub prepare_i()
{
	my $dir= shift;
	my @files;

	# open the directory
	opendir( RB, $dir );
	@files = grep { /^\$I/ } readdir( RB );
	closedir( RB );

	# read all entries
	foreach( @files )
	{
		parse_i_file( $dir . '/' . $_ );
	}
	close(RF);

	# reset the cur_line
	$cur_line = 0;

	return 1;
}

sub parse_i_file()
{
	my $file = shift;
	my $ofs = 0;
	my ($a,$b);
	my $f_size;
	my %record;
	my @array;

	# check the file size
	$f_size = (stat($file))[7];

	# and now to point to the file
	@array  = split(/\//,$file);
	$record{'name'} = $array[$#array];
	$record{'name'} =~ s/^\$I/\$R/;

	# open the file
	open( IF, $file );
	binmode(IF);

	# read the I file
	$ofs = 8;	# starting offset of any data
	$a = Log2t::BinRead::read_32( \*IF,\$ofs);
	$b = Log2t::BinRead::read_32( \*IF,\$ofs);

	$record{'size'} = Log2t::Numbers::join_numbers($a,$b);

	# read the next 8 bytes (the date)
	$a = Log2t::BinRead::read_32( \*IF,\$ofs);
	$b = Log2t::BinRead::read_32( \*IF,\$ofs);

	$record{'date'} =  Log2t::Time::Win2Unix( $a, $b );

	# the path
	$record{'path'} = Log2t::BinRead::read_unicode_end( \*IF, \$ofs, $f_size );

	close( IF );


	$records{$cur_line} = \%record;
	$cur_line++;

	return 1;
}

sub prepare_info2()
{
	# define some variables
	my $return = 1;
	my $ofs;
	my $rec_size;
	my $f_size;
	my $temp;
	my $dir = shift;
	my @dates;
	my $tag;
	my $off;
	my @chars;
	my $path;
	my ( $a, $b );
	my %record;

	# populate the possible drives
	my %drive_list = (
		0x00	=> "A",
		0x01	=> "B",
		0x02	=> "C",
		0x03	=> "D",
		0x04	=> "E",
		0x05	=> "F",
		0x06	=> "G",
		0x07	=> "H",
		0x08	=> "I",
		0x09	=> "J",
		0x0A	=> "K",
		0x0B	=> "L",
		0x0C	=> "M",
		0x0D	=> "N",
		0x0E	=> "O",
		0x0F	=> "P",
		0x10	=> "Q",
		0x11	=> "R",
		0x12	=> "S",
		0x13	=> "T",
		0x14	=> "U",
		0x15	=> "V",
		0x16	=> "W",
		0x17	=> "X",
		0x18	=> "Y",
		0x19	=> "Z"
	);

	# open the access file and read all the lines
	open( FILE, $dir . '/INFO2' ) or $return = 0;

	# read the content of the file and put it into an array
	binmode( FILE );

	# read the record size
	$ofs = 0xC;
	seek(FILE,$ofs,0 );
	read(FILE,$temp,4);
	($a, $b ) =  unpack( "vv", $temp );
	$rec_size = Log2t::Numbers::join_numbers( $a, $b );
	
	# find the file size
	$f_size = (stat($dir . '/INFO2' ))[7];

	# read the entire file, record by record
	$ofs = 0x10; # the location of the first record
	while( $ofs < ($f_size-$rec_size) ) # read all available records
	{
		# null the record value
		%record = '';

		# find the index number
		seek(FILE,$ofs+0x108,0);
		read(FILE,$temp,4);

		# insert the index number into an array
		( $a , $b ) = unpack( "vv", $temp );
		$record{'index'} = Log2t::Numbers::join_numbers( $a, $b );

		# now let's find the drive that contains the file
		seek(FILE,$ofs+0x10C,0);
		read(FILE,$temp,4);

		# insert the drive letter into an array
		$record{'drive'} = $drive_list{unpack( "v",$temp )};

		# read the date of deletion
		seek(FILE,$ofs+0x110,0);
		read(FILE,$temp,8);

		@dates = unpack("VV", $temp );
		$record{'date'} =  Log2t::Time::Win2Unix( $dates[0], $dates[1] );	

		# read the file size as found inside the record
		seek(FILE,$ofs+0x118,0);
		read(FILE,$temp,4);
	
		$record{'size'} = unpack( "V", $temp );

		# now let's read the name in unicode
		$tag = 1;
		$off = 0;
		while( $tag )
		{
			seek(FILE,$ofs+0x11C+$off,0);
			read(FILE,$temp,2);
	
			if( unpack( "v", $temp ) == 0 )
			{
				# the end of the name (end of line)
				$tag = 0;
			}
			else
			{
				# we are reading each character at a time
				push( @chars, $temp );
			}
			$off += 2;
		}
		# join the characters into a single line
		$path = join( '', @chars );
		# delete unneccesary characters
		$path =~ s/\00//g;
		# reset the chars array
		@chars = '';

		# insert the name into the name array
		$record{'path'} =  $path;

		# increment the offset, that is read the next record
		$ofs += $rec_size;

		# now we need to include this into the records hash
		$records{$cur_line} = \%record;

		$cur_line++;
	}

	# reset the cur_line
	$cur_line = 0;

	return $return;;
}

#       load_line
# A subroutine that checks if we have read all the lines in the arrays that store the needed
# values.  If not, it returns an index into the arrays, otherwise -1 is returned to the 
# main script
#
# @return An index into the arrays that contain parsed information (or a -1 if we've reached 
#       the end of the log file)
sub load_line()
{
	# return an index into the arrays that contain read information
	if( $cur_line < keys( %records ) )
	{
		$cur_line++;
		return 1;
	}
	else
	{
		return 0;
	}
}
#       close_file
# A subroutine that closes the file, after it has been parsed
# @return An integer indicating that the close operation was successful
sub close_file()
{
	close( FILE );	
}

#       parse_line
# This function simply takes as a argument an index into the arrays that contain the
# parsed information and creates an array that is returned to the main script
#
# @param INDEX an integer into the arrays that contain parsed information
# @return Returns a array containing the needed values to print a body file
sub parse_line()
{
	# the timestamp object
	my %t_line;
	my $text;
	my $i = $cur_line-1;

	# different information included in different versions of the recycle bin
	if( $older_version )
	{
		$text = 'nr: Dc' . $records{$i}->{'index'} .' - ' . $records{$i}->{'path'} . ' (drive ' . $records{$i}->{'drive'} . ')';
	}
	else
	{
		$text = $records{$i}->{'path'} . ' <-' . $records{$i}->{'name'};
	}

        # content of array t_line ([optional])
        # %t_line {        #       time
        #               index
        #                       value
        #                       type
        #                       legacy
        #       desc
        #       short
        #       source
        #       sourcetype
        #       version
        #       [notes]
        #       extra
        #               [filename]
        #               [md5]
        #               [mode]
        #               [host]
        #               [user]
        #               [url]
        #               [size]
        #               [...]
        # }

	# let's encode this
	#$text = encode( 'utf-8', $text );
	#$text = decode( 'utf-8', $text );

	# create the t_line variable
	%t_line = (
		'time' => { 0 => { 'value' => $records{$i}->{'date'}, 'type' => 'File deleted', 'legacy' => 15 } },
		'desc' => $text,
		'short' => 'DELETED ' . $records{$i}->{'path'},
		'source' => 'RECBIN',
		'version' => 2,
		'extra' => { 'user' => $username, 'host' => $host, 'size' => $records{$i}->{'size'} }
	);

	$t_line{'sourcetype'} = $older_version ? 'RECYCLER' : '$Recycle.bin';

	return \%t_line;
}

#       get_help
# A simple subroutine that returns a string containing the help 
# message for this particular format file.
# @return A string containing a help file for this format file
sub get_help()
{
	return "This plugin parses the recycle bin directory that either contains an INFO2 file or
several files containing the names of \$R and \$I (Vista, Win7 and later versions).  The input module examines the
directory and parses the input according to the version of recycle bin that is being used.

The INFO2 or I\$ files contain information about deleted files found in the Recycle Bin of a Windows OS.  
There are no requirements to use this plugin as it uses native Perl libraries.

The INFO2 file is usually found in \"DRIVE:RECYCLER\\{SID}\\INFO2\"

The input module is used agains the directory, or against
	\"DRIVE_LETTER:\\RECYCLER\\{SID}
or
	\"DRIVE_LETTER:\\\$Recycle.Bin\\{SID}

This format file accepts the following parameters:
	--host HOST
		Hostname to add to the timeline
	-u|--user USER
		Username of the user that this recycle bin belongs to\n";

}

#       verify
# A subroutine that verifies if we are examining a INFO2 document so it can be further 
# processed.  The correct format is a file which starts with the magic value of 0x0500
# @return An array containing an integer and a string.  The integer indicates a success or failure and the
#       string is the error message (if the file is not correctly formed)
sub verify($$)
{
	my $class = shift;
	my $dir = shift;
	my $detail = shift;

	# define an array to keep
	my %return;
	my $line;
	my @files;

	# start by setting the endian correctly
	Log2t::BinRead::set_endian( LITTLE_E );

	# default values
	$return{'success'} = 0;
	$return{'msg'} = 'Not the correct format';

	# first of all we need to be reading a directory content
	return \%return unless -d $$dir;

	# now to verify that we have either INFO2 or $I
	$older_version = -f "$$dir/INFO2" ? 1: 0;
	
	# now to verify the structure (finally)
	if( $older_version )
	{
		# verify that we have an INFO2 file (since we are currently examining an older version of the recycle bin)
		# open the file (at least try to open it)
		eval
		{
			open( FILE, $$dir . '/INFO2' );
			binmode( FILE );
	
			seek(FILE,0,0);
			read(FILE,$line,2);
	
			close( FILE );
		};
		if ( $@ )
		{
			$return{'success'} = 0;
			$return{'msg'} = "Unable to open file";
		}

		if( unpack("v",$line) eq 5 )
		{
			$return{'success'} = 1;
		}
		else
		{
			$return{'success'} = 0;
			$return{'msg'} = "Not the right magic value";
		}
	}
	else
	{
		# now we do not have a INFO2 file, so this is either a Vista/Win7 or later versions of Windows, or not a recycle bin after all
		opendir( RB, $$dir );
		@files = grep { /^\$I/ } readdir( RB );
		closedir( RB );

		# check for the existance of an $I.... file
		if( $#files ge 0 )
		{
			# check for a file
			open( RF, $$dir . '/' . $files[0] );
			binmode( RF );
			my $ofs = 0;

			# read the header
			$line = Log2t::BinRead::read_32(\*RF,\$ofs );

			if( $line eq 0x01 )
			{
				# the first half of the header is verified, read the second half
				$line = Log2t::BinRead::read_32(\*RF,\$ofs );

				if( $line eq 0x00 )
				{
					$return{'success'} = 1;
					$return{'msg'} = 'Magic file is correct';
				}
				else
				{
					$return{'success'} = 0;
					$return{'msg'} = sprintf "Wrong magic value [0x%x]",$line;
				}
			}
			else
			{
				$return{'success'} = 0;
				$return{'msg'} = sprintf "Wrong magic value [0x%x]",$line;
			}

			close( RF );
		}
		else
		{
			$return{'success'} = 0;
			$return{'msg'} = 'No $I... files found';
		}
	}
	return \%return;
}

1;

