#################################################################################################
#		restore	
#################################################################################################
# this script reads a log file and produces a bodyfile containing the timeline information
# that can be used directly with the script mactime from TSK collection.  
# The specification of the body file can be found here:
#	http://wiki.sleuthkit.org/index.php?title=Body_file
# A small script that reads the rp.log file created during restore point creation
# and prints out information from the file
# 
# Author: Kristinn Gudjonsson
# Version : 0.8
# Date : 19/03/10
#
# Copyright 2009 Kristinn Gudjonsson (kristinn ( a t ) log2timeline (d o t) net)
#
#  This file is part of log2timeline.
#
#    log2timeline is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    log2timeline is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with log2timeline.  If not, see <http://www.gnu.org/licenses/>.
#

package Log2t::input::restore;

use strict;
use Fcntl ':mode'; # for permission reading
use Encode;
use Getopt::Long; # read parameters
use Log2t::Time;
use vars qw($VERSION);

# version number
$VERSION = '0.8';

my @dirs;
my $i = 0;
my $rp_dir;
my $host;
my $dir;

# variable for encoding
my $encoding = 'UTF-8';

#       get_version
# A simple subroutine that returns the version number of the format file
#
# @return A version number
sub get_version()
{
	return $VERSION;
}

#       get_description
# A simple subroutine that returns a string containing a description of 
# the funcionality of the format file. This string is used when a list of
# all available format files is printed out
#
# @return A string containing a description of the format file's functionality
sub get_description()
{
	return "Parse the content of the restore point directory"; 
}

#       prepare_file
# This subroutine prepares the loading by reading the content of the 
# restore point directory and storing it in the global array @dirs. 
#
# @params One parameter is defined, the path to the restore point directory 
# @return An integer is returned to indicate whether the file preparation was 
#       successful or not.
sub prepare_file()
{
	my $return = 1;
	my $class = shift;
	$rp_dir = shift;
	my $path = undef;
	my $msg = undef;

        $host = 'unknown';

        # read options from CMD
        @ARGV = @_;
        GetOptions(
                "host:s"	=>	\$host,
		"path:s"=>\$path,
        );


	# open the directory and read the content
	opendir( IMD, $rp_dir ) or $msg = "Could not open the directory $rp_dir";

	if ( defined $msg )
	{
		print STDERR "[RESTORE] Error while trying to open directory ($msg)\n";
		return 0;
	}

	@dirs = readdir( IMD );
	closedir( IMD);

	# sort the entries to get the order at least almost right
	@dirs = sort( @dirs );

	return $return;;
}

#       load_line
# A subroutine that checks if a directory in the restore point directory begins
# with the name RP, indicating that this is a restore point, and returns the value
# of the directory to the main script (if it is a restore point)
#
# @return A string containing the name of the restore point directory (or a -1 if 
#	we've reached the end of the restore points)
sub load_line()
{
	# check the index of the array
	if( $#dirs > $i-1 )
	{
		# check the directory
		if( $dirs[$i] =~ m/^RP/ )
		{
			$dir = $dirs[$i++];
			return 1;
		}
		else
		{
			$i++;
			return load_line();
		}
	}
	else
	{
		return 0;
	}

}

#       close_file
# A subroutine takes care of "closing" the file or directory.
# Since the script has already read and closed the restore point directory
# there is no need to close anything.
#
# @return An integer indicating that the close operation was successful
sub close_file()
{
	$i = 0;
	return 1;
}

#       parse_line
# This is the main "juice" of the format file.  It takes the name of a restore point 
# directory and parses it to produce an array containing all the needed values to print a 
# body file.
# 
# @param LINE a string containing the name of a restore point directory 
# @return Returns a array containing the needed values to print a body file
sub parse_line()
{
	# timestamp object
	my %t_line;
	my $dev;
	my $inode;
	my $mode;
	my $nlink;
	my $uid;
	my $gid;
	my $rdev;
	my $size;
	my $atime;
	my $mtime;
	my $ctime;
	my $blksize;
	my $blocks;
	my $rpinfo;

	# find information about the file (that is the RP directory itself)
	($dev,$inode,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks) = stat("$rp_dir/$dir");
		
	# next to find information from the restore point itself 
	# read the content of the log file (rp.log)
	$rpinfo = read_rpfile( "$rp_dir/$dir" );
	
	##  JLR - 10/11/2010 - return undef if $rpinfo is undef
	##    fix for non-existant rp.log file
	return undef unless (defined $rpinfo);

        # content of array t_line ([optional])
        # %t_line {        #       time
        #               index
        #                       value
        #                       type
        #                       legacy
        #       desc
        #       short
        #       source
        #       sourcetype
        #       version
        #       [notes]
        #       extra
        #               [filename]
        #               [md5]
        #               [mode]
        #               [host]
        #               [user]
        #               [url]
        #               [size]
        #               [...]
        # }

        # create the t_line variable
        %t_line = (
                'time' => { 0 => { 'value' => $rpinfo->{'date'}, 'type' => 'Created', 'legacy' => 15 } },
                'desc' => 'Restore point ' . $dir . ' created - ' . $rpinfo->{'name'},
                'short' => 'Restore point ' . $dir . ' created',
                'source' => 'RP',
                'sourcetype' => 'Restore Point',
                'version' => 2,
                'extra' => { 'host' => $host, 'uid' => $uid, 'gid' => $gid, 'mode' => $mode, 'size' => $size }
        );

	return \%t_line;
}


#       get_help
# A simple subroutine that returns a string containing the help 
# message for this particular format file.
# @return A string containing a help file for this format file
sub get_help()
{
	return "This format file reads the content of a restore point directory
and parses the rp.log file found inside each restore point directory to find information
about the restore point.  

To use this format file point to the directory that contains the restore point, such as
cd /mnt/analyze/System\ Volume\ Information\_restore{.....}/
log2timeline -f restore .

The format file accepts the following parameter
	-h|--host HOSTNAME\n";

}

#       verify
# A subroutine that verifies if we are examining a restore point directory so it can be further 
# processed.  
# The correct format is a directory that consists of a at least one folder which begins with the letters
# RP followed by a number.  Inside this directory is a file called rp.log which is parsed.
#
# @return An array containing an integer and a string.  The integer indicates a success or failure and the
#       string is the error message (if the file is not correctly formed)
sub verify($$)
{
	# define an array to keep
	my %return;
	my @folders;
	my $found = 0;
	my $rp;
	my @words;
	
	# default values
	$return{'success'} = 0;
	$return{'msg'} = '';

	my $class = shift;
	my $dir = shift;
	my $detail = shift;

	return \%return unless -d $$dir;
	
	# open the file (at least try to open it)
	eval
	{
		opendir( DIR, $$dir );
		@folders = grep {/^RP/ && -d $$dir.'/'.$_} readdir( DIR );
		close( DIR );
	};
	if ( $@ )
	{
		$return{'success'} = 0;
		$return{'msg'} = "Unable to open the restore point directory. ";
	}

	# now we have one line of the file, let's read it and verify
	foreach (@folders)
	{
		next if $found;

		if( /^RP/ )
		{
			$found = 1;
			$rp = $_;
		}
	}

	if( $found )
	{
		# there is a RP directory, see if there is a rp.log file inside
		if( -f $$dir . '/' . $rp . '/rp.log' )	
		{
			# no error
			$return{'success'} = 1;
		}
		else
		{
			$return{'success'} = 0;
			$return{'msg'} .= 'No rp.log file inside restore point, perhaps not a restore point directory?. ';
		}
	}
	else
	{
		$return{'msg'} .= 'Unable to find a restore point directory. 
Are there any restore points in the directory in question? (did you check the registry to see if restore point creation is disabled?) ';
		$return{'success'} = 0;
	}

	return \%return;
}

#	read_rpfile
#
# This function reads a rp (restore point) file and displays the
# name of the restore point as well as the date of creation
#
# @params	rp.log
# @return	array containing the name and date of restore point
sub read_rpfile()
{
	# define variables needed
	my ($rp_file, $r_dir);
	my @name;
	my $nafn;
	my $i;	# a buffer
	my $offset;
	my $tag;
	my @dates;
	my %return;
	
	# the directory is the first argument
	$r_dir = $_[0];
	# and then we have a rp.log file underneath
	$rp_file = "$r_dir/rp.log";

	# read the rp.log file
	##open( FILE, "$rp_file" ) || die("Could not open file: $rp_file");
	##  JLR - 10/11/2010 - change to return undef if can't open file rather than die
	unless (open ( FILE, "$rp_file"))
	{
		print STDERR "Could not open the restore point file: $rp_file\n";
		return undef;
	}
	# this is a binary file
	binmode(FILE);

	# read the name, starts in byte 16
	$offset = 0x10;
	
	# read the name, byte by byte
	$tag = 1;
	while( $tag )
	{
		# go to offset of file (starts in byte 16)
		seek(FILE,$offset,0);
		# read 2 bytes, since we are reading a unicode text
		read(FILE,$i,2);
	
		if( unpack( "v", $i ) == 0 )
		{
			# '00' means the end of the name
			$tag = 0;
		}
		else
		{
			# not yet at the end, let's continue and push the value into the array
			push( @name, $i );
		}
		# increase the offset
		$offset += 2;
	}
	
	# now we have the entire name, let's input it into an array
	$nafn = join('',@name);
	# and remove unnecessary information from it
	$nafn =~ s/\00//g;

	$return{'name'} =  $nafn;

	# read the date value
	seek(FILE,-0x8,2);
	read(FILE,$i,8);
	
	# correct format	
	@dates = unpack("VV",$i);

	# find the actual date, using the date function from ptfinder.pl 
	$return{'date'} = Log2t::Time::Win2Unix($dates[0],$dates[1]);

	# close the file
	close( FILE );
		
	# and now we are ready to return the values found 
	return \%return;
}
	


1;
