# Safari
#
# This script handles Safari History.plist files.  Frankly, most of the heavy
# lifting is done by Brian D. Foy's Mac::PropertyList module (available from
# the CPAN).  This module just reports out using the log2timeline API.
#
# Author: Hal Pomeranz (hal.pomeranz@mandiant.com)
# Version: 0.1
# Date: 2010-12-07
#
# Distributed with and under the same licensing terms as log2timeline
#

package Log2t::input::safari;

use strict;
use Mac::PropertyList;
use Getopt::Long;
use Log2t::Common;
use Log2t::Time;
use Encode;

use vars qw($VERSION);

$VERSION = '0.2';

sub get_version
{
    return "$VERSION";
}

sub get_description
{
    return "Parse the contents of a Safari History.plist file";
}

sub get_help
{
    return "Usage: $0 -f safari ... -- [-u username] [-h hostname]

This plugin parses the content of History.plist, a binary property
list file containing Safari browsing history.  On Mac OS X systems,
this file is typically in /User/<username>/Library/Safari";
}

# Module global variables
my $Data = undef;         # Perl data structure produced from plist file
my $UserName = undef;     # Normally passed via command line (-u)
my $HostName = undef;     # Normally passed via command line (-h)
my $Debug = undef;	  # Normally passed from the engine (command line in front-end)

# Check the preamble of the file to see if we're looking at the right thing.
sub verify
{
    my($class, $file) = @_;
    my $buf = undef;

    my %return = ('success' => 0, 'msg' => 'No such file or directory');
    return \%return unless (-f $$file);

    unless (open(FILE, $$file)) {
	$return{'msg'} = "Cannot open $$file: $!";
	return \%return;
    }

    binmode(FILE);
    read(FILE, $buf, 32);
    close(FILE);

    unless ($buf =~ /^bplist00.*WebHistoryFile/) {
	$return{'msg'} = 'Does not appear to be a History.plist file';
	return \%return;
    }

    $return{'success'} = 1;
    $return{'msg'} = 'Success';
    return \%return;
}

# Use Mac::PropertyList::parse_plist_file() to scarf in the plist file.
# Then convert the object hierarchy returned by parse_plist_file() into
# a Perl data structure using the $objects->as_perl() method.
# 
# Also attempt to figure out the username and hostname-- from the command
# line if possible, but otherwise using get_username_from_path()
#
sub prepare_file
{
    my $class = shift;
    my $file = shift;

    my $objects;

    eval { $objects = Mac::PropertyList::parse_plist_file($file); };
    if( $@ ) {
	print STDERR "[Safari] Error $@\n";
	return 0;
    }
    if( ref($objects) ne 'Mac::PropertyList::dict') {
	print STDERR "[Safari] Error.  Object reference is not of type Mac::PropertyList::dict\n";
	return 0;
    }

    eval { $Data = $objects->as_perl; };
    if( $@ ) {
	print STDERR "[Safari] Error occured.  Error message: $@\n";
   	return(0);
    }

    return(0) unless (ref($$Data{'WebHistoryDates'}));
    
    # Read options from CMD
    @ARGV = @_;
    GetOptions(
	"user:s"=>\$UserName, 
	"host:s"=>\$HostName,
	"debug!"=>\$Debug
    );

    # Try really hard to get a user name
    unless (defined($UserName)) {
	$UserName = Log2t::Common::get_username_from_path($file);
    }

    return 1;   
}

# Return true if there are still additional records to process
sub load_line
{
    return scalar(@{$$Data{'WebHistoryDates'}});
}

# Pop the next record off the $$Data{'WebHistoryDates'} list and report out.
sub parse_line
{
    # New %t_line structure.  Most of the basic information is fixed.
    my %return = ('source' => 'WEBHIST',
		  'sourcetype' => 'Safari history',
		  'version' => 2,
		  'extra' => { 'user' => $UserName, 'host' => $HostName },
	);

    # Get a new history record
    my $ref = shift(@{$$Data{'WebHistoryDates'}});

    # Populate fields of %t_list structure from history record
    $return{'short'} = encode('utf-8', "URL: $$ref{''}" );
    $return{'desc'} = "$$ref{''} ($$ref{'title'}) [Visit Count: $$ref{'visitCount'}]";
    $return{'desc'} .= " (redirected from " . join(', ', @{$$ref{'redirectURLs'}}) . ")"
	if (ref($$ref{'redirectURLs'}));
    $return{'desc'} .= " [Non-GET request]" 
	if ($$ref{'lastVisitWasHTTPNonGet'} eq 'true');
    $return{'desc'} = encode( 'utf-8', $return{'desc'} );

    $return{'time'}{0} = { 'value'  => Log2t::Time::mac2epoch($$ref{'lastVisitedDate'}),
			   'type'   => 'Last visited',
			   'legacy' => 15 };

    return \%return;
}

# Nothing to do here since our plist file was open/closed by parse_plist_file()
# in the verify() routine.
sub close_file
{
    return 1;
}

1;
