#################################################################################################
#		squid
#################################################################################################
# this script is a part of the log2timeline program.
# 
# This is a format file that implements a parser for squid access files.  It parses the file
# and provides the main script with enough information to provide a body file that can be
# used in a timeline analysis
# 
# Author: Kristinn Gudjonsson
# Version : 0.4
# Date : 19/03/10
#
# Copyright 2009-2010 Kristinn Gudjonsson (kristinn ( a t ) log2timeline (d o t) net)
#
#  This file is part of log2timeline.
#
#    log2timeline is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    log2timeline is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with log2timeline.  If not, see <http://www.gnu.org/licenses/>.

package Log2t::input::squid;

use strict;
use Getopt::Long; # read parameters
use vars qw($VERSION);
use Log2t::BinRead;
use Log2t::Common ':binary';


# version number
$VERSION = '0.4';

my $host;
my $line;


# 	get_version
# A simple subroutine that returns the version number of the format file
# @return A version number
sub get_version()
{
	return $VERSION;
}

# 	get_description
# A simple subroutine that returns a string containing a description of 
# the funcionality of the format file. This string is used when a list of
# all available format files is printed out
# @return A string containing a description of the format file's functionality
sub get_description()
{
	return 'Parse the content of a Squid access log (http_emulate off)';
}

#	prepare_file
# This subroutine prepares the log file.  It opens the log file and gives the 
# script a handle to the file for further processing.
# @params One parameter is defined, the name and path of the log file to be 
#	parsed.
# @return An integer is returned to indicate whether the file preparation was 
#	successful or not.
sub prepare_file()
{
	my $return = 1;
	my $class = shift;
	my $file = shift;
	my $path;

        $host = 'unknown';

        # read options from CMD
        @ARGV = @_;
        GetOptions(
                "host:s"        =>\$host,
		"path:s"=>\$path,
        );

	# open the access file and read all the lines
	open( FILE, $file ) or $return = 0;

	return $return;;
}

#	load_line
# A subroutine that reads a line from the access file and returns it to the
# main script
# @return A string containing one line of the log file (or a -1 if we've reached 
#	the end of the log file)
sub load_line()
{
	if( $line = <FILE> )
	{
		return 1;
	}
	else
	{
		return 0;
	}
}

#	close_file
# A subroutine that closes the file, after it has been parsed
# @return An integer indicating that the close operation was successful
sub close_file()
{
	close( FILE );	
	return 1;
}

#	parse_line
# This is the main "juice" of the format file.  It takes a line from the log file
# and parses it to produce an array containing all the needed values to print a 
# body file.
# 
# The default structure of Squid log file is:
# timestamp elapsed IP/Client Action/Code Size Method URI Ident Hierarchy/From Content
#
# @param LINE a string containing a single line from the access file
# @return Returns a array containing the needed values to print a body file
sub parse_line()
{
	# log file variables
	my $timestamp;
	my $ip;
	my $action;
	my $size;
	my $method;
	my $uri;
	my $ident;
	my $from;
	my $content;
	my $elapsed;
	my @date; 

	# timestamp object
	my %t_line;

	return 0 unless $line;
	
	$line =~ s/\s+/ /g;
	
	# split the string into variables
	( $timestamp, $elapsed, $ip, $action, $size, $method, $uri, $ident, $from, $content ) = split( / /, $line );

	# fix the timestamp variable
	@date = split( /\./, $timestamp );

        # content of array t_line ([optional])
        # %t_line {        #       time
        #               index
        #                       value
        #                       type
        #                       legacy
        #       desc
        #       short
        #       source
        #       sourcetype
        #       version
        #       [notes]
        #       extra
        #               [filename]
        #               [md5]
        #               [mode]
        #               [host]
        #               [user]
        #               [url]
        #               [size]
        #               [...]
        # }

        # create the t_line variable
        %t_line = (
                'time' => { 0 => { 'value' => $date[0], 'type' => 'Entry written', 'legacy' => 15 } },
                'desc' => $ip . ' connected to \'' . $uri . '\' using ' . $method . ' [' . $action . '] from ' . $from . ', content ' . $content,
                'short' => $ip . " connected to '$uri'",
                'source' => 'NET',
                'sourcetype' => 'Squid access log',
                'version' => 2,
                'extra' => { 'user' => $ip, 'host' => $host, 'src-ip' => $ip, 'size' => $size }
        );

	return \%t_line;
}

#	get_help
# A simple subroutine that returns a string containing the help 
# message for this particular format file.
# @return A string containing a help file for this format file
sub get_help()
{
	return "----------------------------------------------------
	SQUID ACCESS LOG PARSER
----------------------------------------------------
Read squid access logs with the emulate_httpd_log off - use the script
with the FILE option as access.log file\n
\t$0 -f squid access.log
Format of the Squid access file is:
timestamp elapsed IP/Client Action/Code Size Method URI Ident Hierarchy/From Content

The format file accepts one option
	--host HOST\n";

}

#	verify
# A subroutine that reads a single line from the log file and verifies that it is of the
# correct format so it can be further processed.
# The correct format of a Squid access file (with httpd_emulate equal to off) is:
# timestamp elapsed IP/Client Action/Code Size Method URI Ident Hierarchy/From Content
# @return An array containing an integer and a string.  The integer indicates a success or failure and the
#	string is the error message (if the file is not correctly formed)
sub verify($$)
{
	# define an array to keep
	my %return;
	my $line;
	my @words;

	my $class = shift;
	my $file = shift;
	my $detail = shift;

	# default values
	$return{'success'} = 0;
	$return{'msg'} = 'success';

	return \%return unless -f $$file;

	my $ofs = 0;
        # start by setting the endian correctly
        Log2t::BinRead::set_endian( LITTLE_E );

	# open the file (at least try to open it)
	eval
	{
		open( FILE, $$file );
		binmode( FILE );

		unless( $detail )	
		{
			# a line should start with a number, let's verify
			seek(FILE,0,0);
			read(FILE,$line,1);
			$return{'msg'} = 'Wrong magic value';

			if( $line !~ m/[0-9]/ )
			{
				close(FILE);
				return \%return; 
			}
		}

		$line = Log2t::BinRead::read_ascii_until( \*FILE, \$ofs, "\n", 200 );
		close( FILE );
	};
	if ( $@ )
	{
		$return{'success'} = 0;
		$return{'msg'} = "Unable to open file ($@)";
	}
	# now we have one line of the file, let's read it and verify
	# remove unneeded spaces
	$line =~ s/\s+/ /g;
	@words = split(/\s/, $line );
	
	# word count should be 9
	if( $#words eq 9 )
	{
		# verify one variable in the log file, the IP address
		if( $words[2] =~ m/\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/ )
		{
			# the IP address is correctly formed, let's assume other fields are too
			$return{'success'} = 1;
		}
		else
		{
			$return{'error'} = 'IP address field [' .$words[2] . "] not correctly formatted\n";
			$return{'success'} = 0;
		}
	}
	else
	{
		$return{'error'} = 'There should be 9 words per line, instead there are ' . "$#words\n";
		$return{'success'} = 0;
	}

	return \%return;
}

1;
